#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <malloc.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#include "burp.h"

static char *g_programP        = 0;
static char *g_scriptP         = 0;
static char *g_productP        = 0;
static char *g_versionP        = 0;

static int  g_file_memory_size = 0;
static char *g_file_memoryP    = 0;
static int  g_file_lth         = 0;

typedef struct {
  char *m_nameP;
  char *m_valueP;
} overrideT;

/* Overrides passed in on the command line */

static overrideT *g_overridesP = 0;

/* Read the entire input into memory */

static void
remove_comments(void)
{
  char *P, *P1, *P2;
  int  c, in_markup;

  in_markup = 0;
  for (P = P1 = g_file_memoryP; c = *P; ++P) {
    if (!strncmp(P, "<!--", 4)) {
      for (P += 4; strncmp(P, "-->", 3); ++P) {
        if (!*P) {
          fprintf(stderr, "%s: %s has illegally formed comment <!-- ... EOF\n", g_programP, g_scriptP);
          exit(1);
      } }
      P += 2;
      continue;
    }
    switch(c) {
    case '<':
      /* Remove white space before < */
      for (P2 = P1; --P2 >= g_file_memoryP; ) {
        if (!isspace(*P2)) {
          break;
        }
      } 
      P1 = ++P2;
      in_markup = 1;
      break;
    case '>':
      /* Remove white space after > */
      for (P2 = P; isspace(*++P2););
      P = P2 - 1;
      /* Remove white space before > */
      for (P2 = P1; --P2 >= g_file_memoryP; ) {
        if (!isspace(*P2)) {
          break;
        }
      }
      P1 = ++P2;
      in_markup = 0;
      break;
    }
    /* Remove multiple spaces in markup */
    if (in_markup && isspace(c) && P1 > g_file_memoryP && isspace(P1[-1])) {
      continue;
    }
    *P1++ = c;
  }
  *P1 = 0;
  g_file_lth = P1 - g_file_memoryP;
}

#ifndef SSIZE_MAX
#define SSIZE_MAX 0x7FFFFFFF
#endif

static void
read_xml(void)
{
  struct stat buf;
  char        *P, *endP;
  int         lth, segment;
  int         fileno, ret;

  fileno = open(g_scriptP, O_RDONLY);
  if (fileno < 0) {
    fprintf(stderr, "%s: Can't read \"%s\"\n", g_programP, g_scriptP);
    exit(1);
  }

  ret = fstat(fileno, &buf);
  if (ret != 0) {
    fprintf(stderr,"%s: fstat %s returned %d\n", g_programP, g_scriptP, ret);
    perror("");
    exit(1);
  }

  /* Allocate a large enough memory buffer */

  lth = buf.st_size;
  if (!lth) {
    fprintf(stderr, "%s: %s is empty\n", g_programP, g_scriptP);
    exit(1);
  }

  g_file_memory_size = ((lth + 2) + 2047) & ~2047;
  g_file_memoryP     = (char *) malloc(g_file_memory_size);
  if (!g_file_memoryP) {
    fprintf(stderr, "%s: Malloc failed\n", g_programP);
    exit(1);
  }

  P = g_file_memoryP;
  for (endP = P + lth; P < endP; P += ret) {
    segment = endP - P;
    if (segment > SSIZE_MAX) {
      segment = SSIZE_MAX;
    }
    ret = read(fileno, P, segment);
    if (ret <= 0) {
      fprintf(stderr, "%s: read %s returned %d\n", g_programP, g_scriptP, ret);
      perror("");
      exit(1);
  } }
  assert(P == endP);
  if (P[-1] != '\n') {
    *endP++ = '\n';
  }
  *endP = 0;
  g_file_lth = endP - g_file_memoryP;

  close(fileno);
}

burpT	g_output = {0, 0, 0};

static int
print_bash_variables(void)
{
  int	lth, numeric;
  int   product_lth, version_lth;
  char  *start_productP;
  char  *start_versionP;
  int   c, c1, delimiter, depth;
  char  *P, *P2, *P3;
  overrideT * overrideP;
  char  *nameP, *valueP;

  if (strncmp(g_file_memoryP, "<prowl>", 7)) {
    goto illegal;
  }

  depth = 0;
  start_productP = 0;
  start_versionP = 0;
  product_lth    = -1;
  if (g_versionP) {
    version_lth = strlen(g_versionP);
  }
  for (P = g_file_memoryP; c = *P; ++P) {
    switch (c) {
    case '<':
      if (P[1] == '/') {
        for (; (c = *P) != '>'; ++P) {
          if (!c) {
            goto illegal;
        } }
        --depth;
        if (start_versionP && depth < 3) {
          goto done;
        }
        continue;
      }
      /* Start of entity */
      ++depth;
      switch (depth) {
      case 2:
        if (!g_productP) {
          for (P2 = P+1; ; ++P2) {
		    c = *P2;
            if (!c) {
              goto illegal;
            }
            if (isspace(c) || c == '>') {
              break;
		  } }
          product_lth = P2 - P;
		  g_productP  = (char *) malloc(product_lth);
          --product_lth;
          memcpy(g_productP, P+1, product_lth);
          g_productP[product_lth] = 0;
          P = P2;
        } else {
		  if (product_lth < 0) {
            product_lth = strlen(g_productP);
          }
          if (strncmp(P+1, g_productP, product_lth)) {
          	continue;
          }
          P += product_lth + 1;
          if (!isspace(*P) && *P != '>') {
            continue;
        } }
        /* Found correct product section */
        start_productP = P;
        for (; isspace(*P); ++P);
        if (!g_versionP) {
          if (!strncmp(P, "version=", 8)) {
            P += 8;
            delimiter = 0;
            for (; isspace(*P); ++P);
            switch (*P) {
            case '\'':
            case '"':
              delimiter = *P;
              ++P;
            }
            for (P2 = P; *P2 != delimiter && *P2 != '>'; ++P2) {
              if (!c) {
                goto illegal;
            } }
            version_lth  = P2-P;
            g_versionP = (char *) malloc(version_lth + 1);
            strncpy(g_versionP, P, version_lth);
            g_versionP[version_lth] = 0;
            P = P2;
        } }
        break;
      case 3:
        if (!start_productP) {
          continue;
        }
        if (!g_versionP) {
          for (; ; ++P) {
            c = *P;
            if (!c) {
              goto illegal;
            }
            if (isspace(c) || c == '>') {
			  break;
		  } }
        } else {
          if (strncmp(P+1, g_versionP, version_lth)) {
            continue;
          }
          P += version_lth + 1;
          if (*P != '>' && !isspace(*P)) {
            continue;
        } }
        start_versionP = P;
        break;
      case 4:
        if (!start_versionP) {
          continue;
        }
		/* Print name of variable */
        for (P2 = ++P; (c = *P2) != '>' && !isspace(c); ++P2) {
          if (!c) {
            goto illegal;
          }
          fputc(c, stdout);
        }
	    printf("=");
        *P2 = 0;
		g_output.m_lth = 0;

        for (overrideP = g_overridesP; nameP = overrideP->m_nameP; ++overrideP) {
          if (!strcmp(P, nameP)) {
            for (valueP = overrideP->m_valueP; c1 = *valueP; ++valueP) {
              if (c1 == '"') {
				burpc(&g_output, '\\');
              }
              burpc(&g_output, c1);
            }
            break;
        } }
		/* Remove early end of string marker */
        *P2 = c;
	    for (; (c = *P2) != '>'; ++P2) {
          if (!c) {
            goto illegal;
        } }
        for (++P2; (c = *P2) != '<'; ++P2) {
          if (!c) {
            goto illegal;
          }
          switch (c) {
          case '"':
            if (!nameP) {
			  burpc(&g_output, '\\');
            }
            break;
          case '&':
            if (!strncmp(P2,"&amp;",5)) {
              P2 += 4;
              c   = '&';
              break;
            }
            if (!strncmp(P2, "&lt;",4)) {
              P2 += 3;
              c   = '<';
              break;
            }
            if (!strncmp(P2, "&gt;", 4)) {
              P2 += 3;
              c   = '>';
              break;
          } }
          if (!nameP) {
			burpc(&g_output, c);
        } }
			
		if (!g_output.m_lth) {
			numeric = 0;
		} else {
			numeric = 1;
			for (P3 = g_output.m_P; *P3; ++P3) {
				if (isdigit(*P3)) {
					continue;
				}
				if (P3 != g_output.m_P && *P3 == '.' && numeric == 1) {
					++numeric;
					continue;
				}
				numeric = 0;
				break;
		}	}

		if (!numeric) {
        	printf("\"");
		}
		if (g_output.m_lth) {
			fputs(g_output.m_P, stdout);
		}
		if (!numeric) {
			printf("\"");
		}
        printf(";");
        P = P2 - 1;
        break;
  } } }
done:
  return(0);
illegal:
  fprintf(stderr,"%s: %s is not a valid XML configuration file\n", g_programP, g_scriptP);
  return(1);
}

int
main(int argc, char **argv)
{
  overrideT	*overrideP;
  int		i, j;
  char		*P, *P1, *nameP;

  g_overridesP = overrideP = (overrideT *) malloc(sizeof(overrideT) * argc);

  for (i = j = 1; i < argc; ++i) {
    P = argv[i];
    P1 = strchr(P, '=');
    if (!P1 || P1 == P) {
      argv[j++] = P;
    } else {
      overrideP->m_nameP = nameP = (char *) malloc(strlen(P) + 1);
      strcpy(nameP, P);
      nameP += (P1 - P);
      *nameP++ = 0;
      overrideP->m_valueP = nameP;
      ++overrideP;
  } }
  overrideP->m_nameP = 0;
  argc = j;

  switch (argc) {
  case 4:
    g_versionP    = argv[3];
  case 3:
    g_productP = argv[2];
  case 2:
    g_scriptP  = argv[1];
    g_programP = argv[0];
    break;
  default:
    fprintf(stderr, "Prowler version 1.2 " __DATE__ " " __TIME__ "\n");
    fprintf(stderr, "Usage: %s <script.xml> <product> [<version>] [<name>=<value>]*\n", argv[0]);
    exit(1);
  }

  read_xml();
  remove_comments();

  return print_bash_variables();
}
